#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         tables.py
# Purpose:      Helper class to handle tables such as the result table or
#               tle tables
#
#
#
# Author:       p4adch (cadam)
#
# Created:      04/19/2022
# Copyright:    (c) p4adch 2010-
#
# History:
#
# 19.04.2022
# - file created and basic methods
# fixme: add methods for creating and loading a backup of the result
#  table. Include choice for loading the backup if available and no result table
#  is available or it was parsed as argument
# -----------------------------------------------------------------------------

""" Modules """
import csv
import datetime
import gc
import logging
import os
import re
from collections import OrderedDict
from datetime import timedelta
from pathlib import Path

import numpy as np
import pandas as pd
from packaging.version import Version

# Project modules
from . import base_conf as bc

# -----------------------------------------------------------------------------

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021-2023, CLEOSat group'
__credits__ = ["Eduardo Unda-Sanzana, Jeremy Tregloan-Reed, Christian Adam"]
__license__ = "GPL-3.0 license"
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"

__taskname__ = 'tables'
# -----------------------------------------------------------------------------


# Logging
_log = logging.getLogger(__name__)

frmt = "%Y-%m-%dT%H:%M:%S.%f"


class ObsTables(object):
    """Initialize and validate given input."""

    def __init__(self, config, silent=False, log=_log):
        """ Constructor with default values """
        log.setLevel(logging.getLevelName(log.getEffectiveLevel()))

        self.log = log
        self.silent = silent
        self._obs_info = pd.DataFrame()
        self._obj_info = pd.DataFrame()
        self._sat_info = pd.DataFrame()
        self._roi_info = pd.DataFrame()
        self._roi_data = pd.DataFrame()
        self._ext_oi_info = pd.DataFrame()
        self._ext_oi_data = pd.DataFrame()
        self._glint_mask_info = pd.DataFrame()
        self._glint_mask_data = pd.DataFrame()
        self._faint_trail_info = pd.DataFrame()
        self._faint_trail_data = pd.DataFrame()
        self._vis_info = pd.DataFrame()

        self.def_path = Path(config['WORKING_DIR_PATH']).expanduser().resolve()
        self.def_tbl_name = config['RESULT_TABLE_NAME']
        self.def_roi_name = config['ROI_TABLE_NAME']
        self.def_ext_oi_name = config['EXT_OI_TABLE_NAME']
        self.def_glint_mask_name = config['GLINT_MASK_TABLE_NAME']
        self.def_faint_trail_name = config['FAINT_TRAIL_TABLE_NAME']

        self.res_table_fname = self.def_path / self.def_tbl_name
        self.roi_table_fname = self.def_path / self.def_roi_name
        self.ext_oi_table_fname = self.def_path / self.def_ext_oi_name
        self.glint_mask_table_fname = self.def_path / self.def_glint_mask_name
        self.faint_trail_table_fname = self.def_path / self.def_faint_trail_name

        self.def_cols = bc.DEF_RES_TBL_COL_NAMES
        self.def_col_units = bc.DEF_RES_TBL_COL_UNITS
        self.def_key_transl = bc.DEF_KEY_TRANSLATIONS
        self.res_tbl_data = pd.DataFrame(columns=self.def_cols)
        self.res_tbl_units = OrderedDict(zip(self.def_cols,
                                             self.def_col_units[:len(self.def_cols)]))
        self.res_tbl_unit_df = pd.DataFrame([self.res_tbl_units])

    @property
    def obs_info(self):
        return self._obs_info

    @property
    def faint_trail_data(self):
        return self._faint_trail_data

    @property
    def glint_mask_data(self):
        return self._glint_mask_data

    @property
    def roi_data(self):
        return self._roi_data

    @property
    def ext_oi_data(self):
        return self._ext_oi_data

    @property
    def obj_info(self):
        return self._obj_info

    @property
    def sat_info(self):
        return self._sat_info

    @property
    def vis_data(self):
        return self._vis_info

    def find_tle_file(self, sat_name: str,
                      img_filter: str,
                      src_loc: str):
        """ Check for TLE file location """
        level_up = 0
        datetime_in_folder = ('morning' in src_loc or 'evening' in src_loc)
        filter_in_folder = f"/{img_filter}" in src_loc
        if filter_in_folder:
            level_up = 1
            if datetime_in_folder:
                level_up = 2
        else:
            if datetime_in_folder:
                level_up = 1
        search_path = Path(src_loc).parents[level_up]

        # Get times
        date_obs = self._obj_info['Date-Obs'].values[0]
        obs_date = date_obs.split('-')

        # Get satellite type
        sat_type = self.get_sat_type(sat_name)

        tle_filenames = []
        regex_pattern = rf'(?i)tle_{sat_type}_{obs_date[0]}[-_]{obs_date[1]}[-_]{obs_date[2]}.*\.txt'
        regex = re.compile(regex_pattern, re.IGNORECASE)

        path = os.walk(search_path)
        all_files = []

        for root, dirs, files in path:
            matched_files = [s for s in files if re.search(regex, s)]
            for s in matched_files:
                tle_filenames.append((root, s, Path(os.path.join(root, s))))

            all_files.extend((root, s, Path(os.path.join(root, s))) for s in files)

        def select_tle_file(file_names):
            if len(file_names) == 1:
                return file_names[0]

            idx_arr = np.array([i[1].find('unique') for i in file_names])
            unique_idx = np.where(idx_arr == 0)[0]
            if unique_idx.size > 1:
                self.log.warning("Multiple unique TLE files found. "
                                 "This really should not happen!!! TBW: select")
            elif unique_idx.size == 1:
                return files[unique_idx[0]]

            non_unique_idx = np.where(idx_arr == -1)[0]
            if non_unique_idx.size > 1:
                self.log.warning("Multiple files found. "
                                 "This should not happen!!! TBW: select")
                # TODO: Implement user selection here if required
            else:
                return file_names[non_unique_idx[0]]

        if tle_filenames:
            return select_tle_file(tle_filenames)

        # Fallback to the closest prior TLE file by checking obs_date - 1
        previous_day = datetime.datetime(int(obs_date[0]),
                                         int(obs_date[1]),
                                         int(obs_date[2])) - timedelta(days=1)
        prev_date_pattern = (rf'tle_{sat_type}_'
                             rf'{previous_day.year}[-_]'
                             rf'{previous_day.month:02d}[-_]'
                             rf'{previous_day.day:02d}.*\.txt')

        prev_date_regex = re.compile(prev_date_pattern, re.IGNORECASE)

        fallback_files = [(root, s, path) for root, s, path in all_files if re.search(prev_date_regex, s)]

        if fallback_files:
            return select_tle_file(fallback_files)

        self.log.warning("NO TLE file found!!! "
                         "Check the naming/dates of any TLE if present. Exiting for now!!!")
        return None  # Return None if no suitable file is found

    def find_satellite_in_tle(self, tle_loc, sat_name):
        """

        Parameters
        ----------
        tle_loc
        sat_name

        Returns
        -------

        """
        matches = []
        with open(tle_loc, 'r') as file:
            for line in file:
                if sat_name in line:
                    matches.append(line.strip())  # Add the matching line to the list

        if matches:
            match = matches[0]
            self.log.info(f"    Found match for {sat_name}: {match}")
            return match
        else:
            self.log.warning(f"{sat_name} not found in TLE file!!! "
                             "TBW: Get TLE. "
                             "Exiting for now!!!")
            return None

    # def load_glint_mask_table(self):
    #     """Load table with regions of glints that should be masked and excluded"""
    #     glint_info = self._glint_mask_info
    #
    #     # Files info
    #     fname = self.glint_mask_table_fname
    #     if fname.exists():
    #         if not self.silent:
    #             self.log.info(f'> Read glint mask from: {self.def_glint_mask_name}')
    #         glint_info = pd.read_csv(fname, header=0, delimiter=',')
    #
    #     self._glint_mask_info = glint_info
    #
    # def load_roi_table(self):
    #     """Load the regions of interest info table"""
    #     roi_info = self._roi_info
    #
    #     # Files info
    #     fname = self.roi_table_fname
    #     if fname.exists():
    #         if not self.silent:
    #             self.log.info(f'> Read Region-of-Interest from: {self.def_roi_name}')
    #         roi_info = pd.read_csv(fname, header=0, delimiter=',')
    #
    #     self._roi_info = roi_info
    #
    # def load_ext_oi_table(self):
    #     """Load the extensions of interest info table"""
    #     ext_oi_info = self._ext_oi_info
    #
    #     # Files info
    #     fname = self.ext_oi_table_fname
    #     if fname.exists():
    #         if not self.silent:
    #             self.log.info(f'> Read Extensions-of-Interest from: {self.def_ext_oi_name}')
    #         ext_oi_info = pd.read_csv(fname, header=0, delimiter=',')
    #
    #     self._ext_oi_info = ext_oi_info

    def _load_csv_table(self, data_attribute, file_path, table_name, silent=None):
        """Helper method to load CSV tables with common logic

        Parameters
        ----------
        data_attribute : str
            Name of the attribute to update
        file_path : Path
            The path to the CSV file
        table_name : str
            Display name of the table for logging
        silent : bool, optional
            Override the instance's silent flag

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame
        """
        data = getattr(self, data_attribute)

        if file_path.exists():
            if not (self.silent if silent is None else silent):
                self.log.info(f'> Read {table_name} from: {file_path.name}')
            data = pd.read_csv(file_path, header=0, delimiter=',')

        setattr(self, data_attribute, data)
        return data

    def load_glint_mask_table(self):
        """Load table with regions of glints that should be masked and excluded"""
        return self._load_csv_table(
            '_glint_mask_info',
            self.glint_mask_table_fname,
            'glint mask'
        )

    def load_roi_table(self):
        """Load the regions of interest info table"""
        return self._load_csv_table(
            '_roi_info',
            self.roi_table_fname,
            'Region-of-Interest'
        )

    def load_ext_oi_table(self):
        """Load the extensions of interest info table"""
        return self._load_csv_table(
            '_ext_oi_info',
            self.ext_oi_table_fname,
            'Extensions-of-Interest'
        )

    def load_faint_trail_table(self):
        """Load the faint trail table"""
        return self._load_csv_table(
            '_faint_trail_info',
            self.faint_trail_table_fname,
            'Faint-Trail'
        )

    def load_obs_table(self):
        """Load the observation info table"""
        obs_data = self.res_tbl_data

        fname = self.res_table_fname
        if fname.exists():
            if not self.silent:
                self.log.info(f'> Read observation info from: {self.def_tbl_name}')

            df = pd.read_csv(fname, header=None, delimiter=',')
            columns = df.iloc[0].values

            if len(df) > 1 and self.unit_row_matches(df.iloc[1]):
                obs_data = df.iloc[2:].reset_index(drop=True)
            else:
                obs_data = df.iloc[1:].reset_index(drop=True)

            obs_data.columns = columns

        # else:
        #     self.log.error(f'> {self.def_tbl_name} does not exist!! '
        #                     f'This should not happen.')

        self._obs_info = obs_data

    def unit_row_matches(self, unit_row):
        """Check if the row matches the defined units"""
        return any(cell in self.def_col_units or cell == '' for cell in unit_row)

    # def get_roi(self, fname):
    #     """Get the region of interest"""
    #     roi_info = self._roi_info
    #     if not roi_info.empty:
    #         pos_df = roi_info.copy()
    #
    #         # Convert file name to uppercase
    #         pos_df['File'] = roi_info['File'].str.upper()
    #         fname_upper = fname.upper()
    #         # Check position
    #         pos_mask = pos_df['File'] == fname_upper
    #         pos = np.flatnonzero(pos_mask)
    #
    #         if list(pos):
    #             pos_df = roi_info[pos_mask]
    #             self._roi_data = pos_df
    #
    # def get_ext_oi(self, fname):
    #     """Get extension(s) of interest"""
    #     ext_oi_info = self._ext_oi_info
    #
    #     if not ext_oi_info.empty:
    #         pos_df = ext_oi_info.copy()
    #
    #         # Convert file name to uppercase
    #         pos_df['File'] = ext_oi_info['File'].str.upper()
    #         fname_upper = fname.upper()
    #
    #         # Check position
    #         pos_mask = pos_df['File'] == fname_upper
    #         pos = np.flatnonzero(pos_mask)
    #
    #         if list(pos):
    #             pos_df = ext_oi_info[pos_mask]['HDUs'][pos[0]]
    #             self._ext_oi_data = eval(pos_df)
    #         else:
    #             self._ext_oi_data = []

    def _get_table_data(self, source_attribute, target_attribute, fname, process_func=None):
        """Helper method for retrieving table data based on filename

        Parameters
        ----------
        source_attribute : str
            Name of the attribute containing source data
        target_attribute : str
            Name of the attribute to update with matching data
        fname : str
            Filename to search for
        process_func : callable, optional
            Function to process matched data before assignment

        Returns
        -------
        Any
            The matched and processed data
        """
        source_data = getattr(self, source_attribute)

        # Default empty result
        result = pd.DataFrame() if process_func is None else []

        if not source_data.empty:
            # Convert file name to uppercase for case-insensitive matching
            fname_upper = fname.upper()

            # Filter by filename
            mask = source_data['File'].str.upper() == fname_upper
            matches = source_data[mask]

            if not matches.empty:
                result = matches if process_func is None else process_func(matches)

        # Store result in target attribute
        setattr(self, target_attribute, result)
        return result

    def get_roi(self, fname):
        """Get the region of interest"""
        return self._get_table_data('_roi_info', '_roi_data', fname)

    def get_ext_oi(self, fname):
        """Get extension(s) of interest"""

        def process_ext_oi(matches):
            try:
                return eval(matches['HDUs'].iloc[0])
            except (IndexError, KeyError):
                return []

        return self._get_table_data('_ext_oi_info', '_ext_oi_data', fname, process_ext_oi)

    def _get_hdu_table_data(self, source_attribute, target_attribute, fname, hdu_idx):
        """Helper method for retrieving table data based on filename and HDU index

        Parameters
        ----------
        source_attribute : str
            Name of the attribute containing source data
        target_attribute : str
            Name of the attribute to update with matching data
        fname : str
            Filename to search for
        hdu_idx : int
            HDU index to match

        Returns
        -------
        pd.DataFrame
            The matched data
        """
        source_data = getattr(self, source_attribute)

        # Default empty result
        result = pd.DataFrame()

        if not source_data.empty:
            # Convert file name to uppercase for case-insensitive matching
            fname_upper = fname.upper()

            # Filter by filename and HDU index
            mask = (source_data['File'].str.upper() == fname_upper) & (source_data['HDU'] == hdu_idx)
            matches = source_data[mask]

            if not matches.empty:
                result = matches

        # Store result in target attribute
        setattr(self, target_attribute, result)
        return result

    def get_glint_mask(self, fname, hdu_idx):
        """Get Glint masks

        Parameters
        ----------
        fname : str
            Filename to search for in the glint mask table
        hdu_idx : int
            HDU index to match

        Returns
        -------
        pd.DataFrame
            The matched glint mask data
        """
        return self._get_hdu_table_data(
            '_glint_mask_info',
            '_glint_mask_data',
            fname,
            hdu_idx
        )

    def get_faint_trail(self, fname, hdu_idx):
        """Get faint trail data

        Parameters
        ----------
        fname : str
            Filename to search for in the faint trail table
        hdu_idx : int
            HDU index to match

        Returns
        -------
        pd.DataFrame
            The matched faint trail data
        """
        return self._get_hdu_table_data(
            '_faint_trail_info',
            '_faint_trail_data',
            fname,
            hdu_idx
        )

    # def get_glint_mask(self, fname, hdu_idx):
    #     """Get Glint masks
    #
    #     Parameters
    #     ----------
    #     fname : str
    #         Filename to search for in the glint mask table
    #     hdu_idx : int
    #         HDU index to match
    #
    #     Returns
    #     -------
    #     pd.DataFrame
    #         The matched glint mask data
    #     """
    #     glint_info = self._glint_mask_info
    #
    #     if glint_info.empty:
    #         self._glint_mask_data = pd.DataFrame()
    #         return
    #
    #     # Convert file name to uppercase for case-insensitive matching
    #     fname_upper = fname.upper()
    #
    #     # Filter by filename then by HDU index (more efficient than making a copy)
    #     mask = (glint_info['File'].str.upper() == fname_upper) & (glint_info['HDU'] == hdu_idx)
    #     matches = glint_info[mask]
    #
    #     # Store matched data
    #     if not matches.empty:
    #         self._glint_mask_data = matches
    #     else:
    #         self._glint_mask_data = pd.DataFrame()
    #
    # def get_faint_trail(self, fname, hdu_idx):
    #     """Get faint trail data
    #
    #     Parameters
    #     ----------
    #     fname : str
    #         Filename to search for in the faint trail table
    #     hdu_idx : int
    #         HDU index to match
    #
    #     Returns
    #     -------
    #     pd.DataFrame
    #         The matched faint trail data
    #     """
    #     faint_info = self._faint_trail_info
    #
    #     if faint_info.empty:
    #         self._faint_trail_data = pd.DataFrame()
    #         return
    #
    #     # Convert file name to uppercase for case-insensitive matching
    #     fname_upper = fname.upper()
    #
    #     # Filter by filename then by HDU index (more efficient than making a copy)
    #     mask = (faint_info['File'].str.upper() == fname_upper) & (faint_info['HDU'] == hdu_idx)
    #     matches = faint_info[mask]
    #
    #     # Store matched data
    #     if not matches.empty:
    #         self._faint_trail_data = matches
    #     else:
    #         self._faint_trail_data = pd.DataFrame()

    # def get_glint_mask(self, fname, hdu_idx):
    #     """Get Glint masks"""
    #
    #     glint_info = self._glint_mask_info
    #
    #     if not glint_info.empty:
    #         pos_df = glint_info.copy()
    #
    #         # Convert file name to uppercase
    #         pos_df['File'] = glint_info['File'].str.upper()
    #         fname_upper = fname.upper()
    #
    #         # Check position
    #         pos_mask = (pos_df['File'] == fname_upper)
    #         pos = np.flatnonzero(pos_mask)
    #
    #         if list(pos):
    #             df = glint_info[pos_mask]
    #             hdu_match = (df['HDU'] == hdu_idx)
    #             if np.any(hdu_match):
    #
    #                 if len(np.array(hdu_match)) == 1:
    #                     idx = pos[hdu_match]
    #                     self._glint_mask_data = glint_info.iloc[idx]
    #                 else:
    #                     idx = pos[np.where(hdu_match)[0]]
    #                     self._glint_mask_data = glint_info.iloc[idx]

    def get_object_data(self, fname, kwargs, obsparams, hdu_idx):
        """
        Extract a row from the observation info table.

        Parameters
        ----------
        fname : str
            The name of the file to extract data from.
        kwargs : dict
            Additional keyword arguments containing observation parameters.
        obsparams : dict
            Dictionary containing observation parameters.
        hdu_idx : int
            Index of the HDU (Header Data Unit).

        Returns
        -------
        None
        """
        obs_info = self._obs_info
        n_ext = obsparams['n_ext']

        # Make a copy
        obs_info_tmp = obs_info.copy()

        # Convert file name to uppercase
        obs_info_tmp['File'] = obs_info['File'].str.upper()
        fname_upper = fname.upper()

        # Check position
        mask = obs_info_tmp['File'] == fname_upper
        pos = np.flatnonzero(mask)

        ra_key, ra_val = self.get_position_data(kwargs, obsparams, 'ra', 'RA')
        de_key, de_val = self.get_position_data(kwargs, obsparams, 'dec', 'DEC')
        inst_key, inst_val = self.get_position_data(kwargs, obsparams, 'instrume', 'Instrument')
        obj_key, obj_val = self.get_position_data(kwargs, obsparams, 'object', 'Object')

        key_list = ['RA', 'DEC', 'Instrument', 'Object']
        key_dict = {'RA': [ra_val], 'DEC': [de_val], 'Instrument': [inst_val],
                    'Object': [obj_val]}

        if n_ext > 1:
            key_list = ['RA', 'DEC', 'Instrument', 'Object', 'HDU_idx']
            key_dict['HDU_idx'] = [f"{hdu_idx:.1f}"]

        if list(pos):
            self.log.debug("  Possible match with previous entry found. Check RA and DEC.")

            df = obs_info[mask][key_list].astype(str)

            coord_match = df.isin(key_dict).all(axis='columns')

            if coord_match.any():
                if len(coord_match) == 1:
                    self.log.debug("  Coordinate/Pointing match found. Updating.")
                    idx = pos[coord_match]
                    self._obj_info = obs_info.iloc[idx]
                else:
                    self.log.debug("  Multiple Coordinate/Pointing matches found."
                                   " This should not happen.")
                    idx = pos[np.where(coord_match)[0]]
                    self._obj_info = obs_info.iloc[idx]
            else:
                self.log.error("  NO Coordinate/Pointing match found."
                               " This should not happen!!!")
        else:
            self.log.error("  File not found. This should not happen!!!")

    def get_satellite_id(self, obj_in):
        """
        Get satellite id from object name.

        This function takes an object name, normalizes it, and determines the satellite type and unique ID if present.

        Parameters
        ----------
        obj_in : str
            The input object name.

        Returns
        -------
        tuple
            A tuple containing the satellite identifier and a unique ID (if present).
        """
        self.log.debug("  Identify Satellite/Object name")

        # Standardize object name: uppercase and replace spaces with hyphens
        obj_name = obj_in.upper().replace(' ', '-')

        # Extract unique ID if present (e.g., "-IDxxxx")
        unique_id = None
        if '-ID' in obj_name:
            idx = obj_name.index('-ID')
            unique_id = obj_name[idx + 1:idx + 8]
            obj_name = obj_name[:idx]

        # Remove extensions like "-N" if present
        if '-N' in obj_name:
            idx = obj_name.index('-N')
            obj_name = obj_name[:idx]

        # Remove binning identifiers (e.g., "_1X1", "-2X2")
        for suffix in ['_1X1', '-1X1', '_2X2', '-2X2', '_4X4', '-4X4']:
            if suffix in obj_name:
                obj_name = obj_name.replace(suffix, '')

        # Replace remaining underscores with hyphens
        obj_name = obj_name.replace('_', '-')

        # Identify satellite type
        sat_name = obj_name
        found = False
        for key, props in bc.SATELLITE_PROPERTIES.items():
            if any(identifier in obj_name for identifier in props['identifiers']):
                sat_name = key
                found = True
                break

        # Extract satellite number if present and format accordingly
        parts = obj_name.split('-')
        nb_str = None
        for part in parts:
            if part.isdigit():
                nb_str = part
                break

        if found:
            if nb_str:
                sat_id = bc.SATELLITE_PROPERTIES[sat_name]['format_number'](nb_str)
            else:
                sat_id = sat_name
        else:
            # If not a known satellite observation, return the original object name without modifications
            sat_id = obj_name

        return sat_id, unique_id

    def get_position_data(self, kwargs, obsparams, obskey, dkey):
        """

        Parameters
        ----------
        kwargs
        obsparams
        obskey
        dkey

        Returns
        -------

        """
        key_transl = self.def_key_transl
        for i in range(len(key_transl[dkey])):
            key = key_transl[dkey][i]
            if key in kwargs and key == obsparams[obskey]:
                val = str(kwargs[key])
                return key, val

    def update_obs_table(self, file, kwargs, obsparams):
        """
        Load the observation obs info table.

        Load the table with observation information.
        If the file does not exist, it is created.
        """
        obs_info = self._obs_info
        fname = self.res_table_fname
        n_ext = obsparams['n_ext']

        # exposure time
        expt = kwargs['EXPTIME']

        # Update the columns in case the table was changed
        for col in self.def_cols:
            if col not in obs_info:
                # insert column at the end
                obs_info.insert(len(obs_info.columns), col, '')

        # Reorder columns
        obs_info = obs_info[self.def_cols]

        # Replace all None values with NAN
        obs_info = obs_info.replace(to_replace=[None], value=np.nan, inplace=False)

        # Make a copy
        obs_info_tmp = obs_info.copy()

        # Convert file name to uppercase
        obs_info_tmp['File'] = obs_info['File'].str.upper()
        file_upper = file.upper()

        # Check the position of data
        mask = obs_info_tmp['File'] == file_upper
        pos = np.flatnonzero(mask)

        ra_key, ra_val = self.get_position_data(kwargs, obsparams, 'ra', 'RA')
        de_key, de_val = self.get_position_data(kwargs, obsparams, 'dec', 'DEC')
        inst_key, inst_val = self.get_position_data(kwargs, obsparams, 'instrume', 'Instrument')
        obj_key, obj_val = self.get_position_data(kwargs, obsparams, 'object', 'Object')

        key_list = ['RA', 'DEC', 'Instrument', 'Object']
        key_dict = {'RA': [ra_val], 'DEC': [de_val], 'Instrument': [inst_val],
                    'Object': [obj_val]}

        if n_ext > 1:
            hdu_idx = str(float(kwargs['HDU_IDX']))
            key_list = ['RA', 'DEC', 'Instrument', 'Object', 'HDU_idx']
            key_dict['HDU_idx'] = [hdu_idx]

        if not list(pos):
            self.log.debug("  File name not found. Adding new entry.")
            obs_info = self.add_row(obs_info, file, kwargs, self.def_cols,
                                    expt, obj_key, ra_val, de_val)
        else:
            self.log.debug("  Possible match with previous entry found. "
                           "Check RA, DEC, object and instrument.")

            df = obs_info[mask][key_list].astype(str)

            coord_match = df.isin(key_dict).all(axis=1)
            if coord_match.any():

                self.log.debug("  Coordinate/Pointing, object and instrument match found. Updating.")
                idx = pos[np.where(coord_match)[0]]
                obs_info = self.update_row(obs_info, idx, kwargs, self.def_cols,
                                           expt, obj_key, ra_val, de_val)

            else:
                self.log.debug("  NO Coordinate/Pointing, object and instrument match found. Adding entry.")
                obs_info = self.add_row(obs_info, file, kwargs,
                                        self.def_cols, expt, obj_key, ra_val, de_val)

        # Check for duplicates
        bool_series = obs_info.duplicated(keep='first')
        obs_info = obs_info[~bool_series]

        obs_info = obs_info.sort_values(by=['Date-Obs', 'Obs-Start', 'File'],
                                        ascending=[True, True, True]).reset_index(drop=True)

        self._obs_info = obs_info
        self.save_table(fname, obs_info)

        del obs_info, obs_info_tmp
        gc.collect()

    def update_row(self, obs_info, idx, kwargs, def_cols, expt, obj_key, ra, dec):
        """
        Update row in result table.

        This function updates a specific row in the result table with the provided
        observation information and parameters.
        """

        for i in range(1, len(def_cols)):
            dcol_name = def_cols[i]
            for key, value in kwargs.items():
                if dcol_name in self.def_key_transl and \
                        key in self.def_key_transl[dcol_name]:
                    if dcol_name == 'Object':
                        # Create uniform satellite id from object name
                        sat_name, alt_id = self.get_satellite_id(kwargs[obj_key])
                        obs_info.loc[idx, dcol_name] = value
                        obs_info.loc[idx, 'Sat-Name'] = sat_name
                        obs_info.loc[idx, 'UniqueID'] = alt_id
                    elif dcol_name == 'Date-Obs':
                        t_short, t_start, t_mid, t_stop = self.update_times(value, expt, kwargs)
                        obs_info.loc[idx, dcol_name] = t_short
                        obs_info.loc[idx, 'Obs-Start'] = t_start
                        obs_info.loc[idx, 'Obs-Mid'] = t_mid.strftime('%H:%M:%S.%f')[:-3]
                        obs_info.loc[idx, 'Obs-Stop'] = t_stop.strftime('%H:%M:%S.%f')[:-3]
                    elif dcol_name == 'WCS_cal':
                        obs_info.loc[idx, dcol_name] = 'T' if value or value == 'successful' else 'F'
                    elif dcol_name == 'RA':
                        obs_info.loc[idx, dcol_name] = ra
                    elif dcol_name == 'DEC':
                        obs_info.loc[idx, dcol_name] = dec
                    else:
                        obs_info.loc[idx, dcol_name] = value
                else:
                    if key == dcol_name and dcol_name not in ['Obs-Start', 'Obs-Mid', 'Obs-Stop', 'Sat-Name']:
                        obs_info.loc[idx, key] = value

        return obs_info

    def add_row(self, obs_info, file, kwargs, def_cols, expt, obj_key, ra, dec):
        """ Add new row to result table"""
        new_dict = {def_cols[0]: [file]}
        for i in range(len(def_cols)):
            dcol_name = def_cols[i]
            for key, value in kwargs.items():
                if dcol_name in self.def_key_transl and \
                        key in self.def_key_transl[dcol_name]:
                    if dcol_name == 'Object':
                        # Create uniform satellite id from object name
                        sat_name, alt_id = self.get_satellite_id(kwargs[obj_key])
                        new_dict[dcol_name] = value
                        new_dict['Sat-Name'] = sat_name
                        new_dict['UniqueID'] = alt_id
                    elif dcol_name == 'Date-Obs':
                        t_short, t_start, t_mid, t_stop = self.update_times(value, expt, kwargs)
                        new_dict[dcol_name] = t_short
                        new_dict['Obs-Start'] = t_start
                        new_dict['Obs-Mid'] = t_mid.strftime('%H:%M:%S.%f')[:-3]
                        new_dict['Obs-Stop'] = t_stop.strftime('%H:%M:%S.%f')[:-3]
                    elif dcol_name == 'WCS_cal':
                        new_dict[dcol_name] = 'T' if value or value == 'successful' else 'F'
                    elif dcol_name == 'RA':
                        new_dict[dcol_name] = ra
                    elif dcol_name == 'DEC':
                        new_dict[dcol_name] = dec
                    else:
                        new_dict[dcol_name] = value
                else:
                    if key == dcol_name and dcol_name not in ['Obs-Start', 'Obs-Mid', 'Obs-Stop', 'Sat-Name']:
                        new_dict[dcol_name] = value

        new_df = pd.DataFrame(new_dict)

        # Pandas version < 2.0
        if Version(pd.__version__) < Version('2.0'):
            obs_info = obs_info.append(new_df, ignore_index=False)
        else:
            obs_info = pd.concat([obs_info, new_df], ignore_index=False)

        del new_df, new_dict
        gc.collect()

        return obs_info

    def read_vis_file(self, vis_file_list: list):
        """Read the visibility file"""

        col_names = bc.DEF_VIS_TBL_COL_NAMES
        df_list = []
        for i in range(len(vis_file_list)):
            vis_loc = vis_file_list[i]
            file = vis_loc[1]

            # Create 3 empty lists to hold row data
            data, data2, data3 = ([] for _ in range(3))

            # Load the .csv into a list
            with open(file, 'r') as f:
                reader = csv.reader(f)
                lines = list(reader)

            # Check if the first line has multiple words
            has_multiple_words = self.contains_multiple_words(lines[1][0])
            two_lines = False if has_multiple_words else True

            for line in lines[1:]:

                # Split line using regex
                row = re.split(pattern=r'\s\s|\s|\t', string=line[0])
                l_row = len(row)

                if two_lines:

                    if l_row in [1, 2]:
                        test_str = row[0] if l_row == 1 else '-'.join(row)
                        sat_id, alt_id, unique_id = self.identify_satellite(test_str)
                        data2.append([sat_id, alt_id, unique_id])
                    if l_row == 13:
                        data3.append(row)
                else:
                    test_str, rest = (row[0], row[1:]) if l_row == 14 else ('-'.join(row[:2]), row[2:])
                    sat_id, alt_id, unique_id = self.identify_satellite(test_str)
                    tmp = [sat_id, alt_id, unique_id]
                    tmp += rest
                    data.append(tmp)
            if two_lines:
                for j in range(len(data2)):
                    tmp = data2[j]
                    tmp += data3[j]
                    data.append(tmp)

            df = pd.DataFrame(data=data, columns=col_names)
            df.replace(to_replace=[None], value=np.nan, inplace=True)
            df.reset_index()
            df_list.append(df)

            del data, df
            gc.collect()

        df = pd.concat(df_list)
        df.reset_index()

        self._vis_info = df

        del df, df_list
        gc.collect()

    def save_table(self, fname, obs_info):
        """Method to save the result table to a csv file."""
        df_units = self.res_tbl_unit_df
        if obs_info.empty:
            df_data = pd.concat([pd.DataFrame([self.def_cols], columns=self.def_cols),
                                 df_units], ignore_index=True)
        else:
            df_data = pd.concat([pd.DataFrame([self.def_cols], columns=self.def_cols),
                                 df_units, obs_info], axis=0, ignore_index=True)

        df_data.to_csv(fname, index=False, header=False)

    def update_times(self, value, expt, kwargs):
        """ Update Observation date, start, mid, and end times."""
        self.log.debug("  Check Obs-Date")

        tel_key = 'TELESCOP' if 'TELESCOP' in kwargs else 'OBSERVAT'
        if 'TIME-OBS' in kwargs and kwargs[tel_key] != 'CTIO 4.0-m telescope':
            t = pd.to_datetime(f"{kwargs['DATE-OBS']}T{kwargs['TIME-OBS']}",
                               format=frmt, utc=False)
        else:
            t = pd.to_datetime(value, format=frmt, utc=False)

        t_short = t.strftime('%Y-%m-%d')
        t_start = t.strftime('%H:%M:%S.%f')[:-3]
        t_mid = t + timedelta(seconds=float(expt) / 2)
        t_stop = t + timedelta(seconds=float(expt))

        return t_short, t_start, t_mid, t_stop

    @staticmethod
    def contains_multiple_words(s):
        return len(re.compile(r' ').split(s)) > 1

    @staticmethod
    def identify_satellite(test_str):

        unique_id = re.findall(r'(ID-\d+)', test_str)
        unique_id = unique_id[0] if unique_id else None

        if unique_id is not None:
            test_str = test_str.replace(f'-{unique_id}', '')
        alt_id = re.findall(r'-\((\w+)\)', test_str)
        alt_id = alt_id[0] if alt_id else None

        if alt_id is not None:
            test_str = test_str.replace(f'-({alt_id})', '')

        return test_str, alt_id, unique_id

    @staticmethod
    def get_sat_type(sat_name):
        """
        Determine the satellite type based on the satellite name.

        Parameters
        ----------
        sat_name : str
            The name of the satellite.

        Returns
        -------
        str
            The satellite type corresponding to the given satellite name.
            Returns 'oneweb' if no specific match is found.
        """
        sat_type_mapping = {
            'STARLINK': 'starlink',
            'BLUEWALKER': 'bluewalker',
            'KUIPER': 'kuiper',
            'SPACEMOBILE': '[spacemobile|all]'
        }

        # Iterate over the mapping to find a matching satellite type
        for key, value in sat_type_mapping.items():
            if key in sat_name.upper():
                return value

        # Default to 'oneweb' if no match is found
        return 'oneweb'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         photometry.py
# Purpose:      Utilities for the extraction of photometry from the image.
#
#
#
#
# Author:       p4adch (cadam)
#
# Created:      05/09/2022
# Copyright:    (c) p4adch 2010-
#
# History:
#
# 09.05.2022
# - file created and basic methods
#
# -----------------------------------------------------------------------------

""" Modules """
import gc
import os
import sys
import inspect
import logging
from concurrent.futures import ProcessPoolExecutor
from functools import partial

import numpy as np
import pandas as pd

try:
    import matplotlib
    from matplotlib import pyplot as plt
    import matplotlib.lines as mlines

except ImportError:
    plt = None
else:
    import matplotlib
    import matplotlib.gridspec as gridspec  # GRIDSPEC !
    from matplotlib.ticker import (AutoMinorLocator, LogLocator)
    from astropy.visualization import (LinearStretch, LogStretch, SqrtStretch)
    from astropy.visualization.mpl_normalize import ImageNormalize
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    # matplotlib parameter
    matplotlib.use('Qt5Agg')
    matplotlib.rc("lines", linewidth=1.2)
    matplotlib.rc('figure', dpi=150, facecolor='w', edgecolor='k')
    matplotlib.rc('text.latex', preamble=r'\usepackage{sfmath}')

from astropy.stats import sigma_clipped_stats
from photutils.aperture import (aperture_photometry,
                                ApertureStats,
                                CircularAperture,
                                CircularAnnulus,
                                RectangularAperture,
                                RectangularAnnulus)

# pipeline-specific modules
from . import base_conf as bc

# -----------------------------------------------------------------------------

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021-2023, CLEOSat group'
__credits__ = ["Eduardo Unda-Sanzana, Jeremy Tregloan-Reed, Christian Adam"]
__license__ = "GPL-3.0 license"
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"

__taskname__ = 'photometry'
# -----------------------------------------------------------------------------

""" Parameter used in the script """
log = logging.getLogger(__name__)

MODULE_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))


# -----------------------------------------------------------------------------

def convert_sdss_to_BVRI_lupton(f, x1, x2, x3):
    """
    Convert to BVRI magnitudes using SDSS magnitudes.

    Lupton et al.(2005) BibCode: 2005AAS...20713308L

    """
    coeff_dict = bc.CONV_SSDS_BVRI

    y1 = x2[:, 0] + coeff_dict[f][0][0] * (x1[:, 0] - x2[:, 0]) + coeff_dict[f][0][1]
    e_y1 = np.sqrt((coeff_dict[f][0][0] * x1[:, 1]) ** 2 +
                   ((1 - coeff_dict[f][0][0]) * x2[:, 1]) ** 2 +
                   coeff_dict[f][0][2] ** 2)
    y2 = x2[:, 0] + coeff_dict[f][1][0] * (x2[:, 0] - x3[:, 0]) + coeff_dict[f][1][1]
    e_y2 = np.sqrt((coeff_dict[f][1][0] * x3[:, 1]) ** 2 +
                   ((1 - coeff_dict[f][1][0]) * x2[:, 1]) ** 2 +
                   coeff_dict[f][1][2] ** 2)
    if f in ['B', 'I']:
        y1 = x1[:, 0] + coeff_dict[f][0][0] * (x1[:, 0] - x2[:, 0]) + coeff_dict[f][0][1]
        e_y1 = np.sqrt((coeff_dict[f][0][0] * x2[:, 1]) ** 2 +
                       ((1 - coeff_dict[f][0][0]) * x1[:, 1]) ** 2 +
                       coeff_dict[f][0][2] ** 2)

    _mags = np.mean(list(zip(y1, y2)), axis=1)
    e_mags = np.mean(list(zip(e_y1, e_y2)), axis=1)

    del f, x1, x2, x3
    gc.collect()

    return _mags, e_mags

def convert_sdss_to_U_jester(u, g, r):
    """
    Convert to U magnitude using SDSS magnitudes.

    Jester et al.(2005)

    """
    B = g[:, 0] + 0.33 * (g[:, 0] - r[:, 0]) + 0.20
    U = 0.77 * (u[:, 0] - g[:, 0]) - 0.88 + B
    e_U = np.sqrt((0.77 * u[:, 1])**2 +
                  (0.10 * g[:, 1])**2 +
                  (0.33 * r[:, 1])**2)

    return U, e_U

# Define the function that will be executed in parallel
def process_aper_photometry(r_aper, iteration, image, src_pos, mask, aper_mode, fwhm, exp_time,
                            r_in=1.9, r_out=2.2, width=500, theta=0., gain=1., rdnoise=0,
                            total=1, use_lock=False):
    """

    Parameters
    ----------
    r_aper
    iteration
    image
    src_pos
    mask
    aper_mode
    fwhm
    exp_time
    r_in
    r_out
    width
    theta
    gain
    rdnoise
    total
    use_lock

    Returns
    -------

    """
    if aper_mode == 'circ':

        phot_res = get_aper_photometry(image, src_pos, mask=mask,
                                       aper_mode=aper_mode,
                                       r_aper=r_aper * fwhm,
                                       r_in=r_in * fwhm,
                                       r_out=r_out * fwhm)
    else:
        height = r_aper * fwhm * 2.
        w_in = width + r_in * fwhm
        w_out = width + r_out * fwhm
        h_in = r_in * fwhm * 2.
        h_out = r_out * fwhm * 2.
        phot_res = get_aper_photometry(image, src_pos, mask=mask,
                                       height=height,
                                       width=width, w_in=w_in, w_out=w_out, h_in=h_in,
                                       h_out=h_out, theta=theta, aper_mode='rect')

    aper_flux_counts, _, _, aper_bkg_counts, _, area = phot_res

    aper_flux = aper_flux_counts / exp_time
    aper_bkg = aper_bkg_counts / exp_time

    # Calculate the Signal-to-Noise ratio
    snr, snr_err = get_snr(flux_star=aper_flux, flux_bkg=aper_bkg,
                           t_exp=exp_time, area=area,
                           gain=gain, rdnoise=rdnoise, dc=0.)
    del image, mask

    bc.print_progress_bar(iteration, total, length=80,
                          color=bc.BCOLORS.OKGREEN, use_lock=use_lock)

    return aper_flux, aper_bkg, snr, snr_err


def get_opt_aper_trail(image,
                       src_pos,
                       fwhm,
                       exp_time,
                       config: dict,
                       img_mask=None,
                       width: float = 500.,
                       theta: float = 0.,
                       gain: float = 1,
                       rdnoise: float = 0,
                       silent: bool = False):
    """

    Parameters
    ----------
    image
    src_pos
    fwhm
    exp_time
    config
    img_mask
    width
    theta
    gain
    rdnoise
    silent

    Returns
    -------

    """
    # Initialize logging for this user-callable function
    log.setLevel(logging.getLevelName(log.getEffectiveLevel()))

    dstep = config['TRAIL_APER_STEP_SIZE']
    start = config['TRAIL_APER_START']
    stop = config['TRAIL_APER_STOP']
    rapers = np.linspace(start, stop, int((stop - start) / dstep + 1))

    output = np.zeros((len(rapers), len(src_pos), 4))
    output[output == 0] = np.nan

    r_in = config['TRAIL_RSKYIN']
    r_out = config['TRAIL_RSKYOUT']

    total_rapers = len(rapers)
    try:
        # Create a partial function with the common parameters
        func = partial(process_aper_photometry, image=image, src_pos=src_pos,
                       mask=img_mask, fwhm=fwhm, exp_time=exp_time, aper_mode='rect',
                       r_in=r_in, r_out=r_out, width=width, theta=theta,
                       gain=gain, rdnoise=rdnoise, total=total_rapers, use_lock=True)

        # Map the function over the range of rapers and get the results
        with ProcessPoolExecutor() as executor:
            results = list(executor.map(func, rapers, range(1, total_rapers + 1)))

        # Store the results in the output array
        for i, result in enumerate(results):
            # Assuming result is of the form suitable to be assigned to the output array
            output[i, :, 0], output[i, :, 1], output[i, :, 2], output[i, :, 3] = result

        bc.print_progress_bar(total_rapers, total_rapers,  length=80,
                              color=bc.BCOLORS.OKGREEN, use_lock=True)

    finally:
        # Ensure the executor is shut down properly
        executor.shutdown(wait=True)
        sys.stdout.write('\n')

    del executor

    max_snr_idx = np.nanargmax(output[:, :, 2], axis=0)
    max_snr_aprad = float(rapers[max_snr_idx])
    optimum_aprad = max_snr_aprad * 1.25  # increased by 25%
    qlf_aptrail = True

    if not silent:
        log.info('    ==> best-fit aperture height: %3.1f (FWHM)' % (max_snr_aprad * 2))
        log.info('    ==> optimum aperture height (h x 1.25): %3.1f (FWHM)' % (optimum_aprad * 2))

    del image
    return output, rapers, max_snr_aprad, optimum_aprad, qlf_aptrail


def get_optimum_aper_rad(image: np.ndarray,
                         std_cat: pd.DataFrame,
                         fwhm: float,
                         exp_time,
                         config: dict,
                         aper_rad: float = 1.7,
                         r_in: float = 1.9,
                         r_out: float = 2.2,
                         gain: float = 1,
                         rdnoise: float = 0,
                         silent: bool = False):
    """

    Parameters
    ----------
    image
    std_cat
    fwhm
    exp_time
    config
    aper_rad
    r_in
    r_out
    gain
    rdnoise
    silent

    Returns
    -------

    """
    # Initialize logging for this user-callable function
    log.setLevel(logging.getLevelName(log.getEffectiveLevel()))

    # Mask unwanted pixel
    mask = (image < 0)
    img_mask = config['img_mask']
    # if img_mask is not None:
    #     mask |= img_mask

    dstep = config['APER_STEP_SIZE']
    start = config['APER_START']
    stop = config['APER_STOP']
    rapers = np.linspace(start, stop, int((stop - start) / dstep + 1))

    # Select only the 25 brightest sources
    std_cat = std_cat.head(25)

    # Get positions
    src_pos = np.array(list(zip(std_cat['xcentroid'], std_cat['ycentroid'])))

    output = np.zeros((len(rapers), len(src_pos), 4))
    output[output == 0] = np.nan

    total_rapers = len(rapers)
    try:
        # Create a partial function with the common parameters
        func = partial(process_aper_photometry, fwhm=fwhm,
                       exp_time=exp_time, gain=gain, rdnoise=rdnoise, image=image,
                       src_pos=src_pos, mask=img_mask, r_in=r_in, r_out=r_out,
                       aper_mode='circ', total=total_rapers, use_lock=True)

        # Map the function over the range of rapers and get the results
        with ProcessPoolExecutor() as executor:
            results = list(executor.map(func, rapers, range(1, total_rapers + 1)))

        # Store the results in the output array
        for i, result in enumerate(results):
            output[i, :, 0], output[i, :, 1], output[i, :, 2], output[i, :, 3] = result

        bc.print_progress_bar(total_rapers, total_rapers,  length=80,
                              color=bc.BCOLORS.OKGREEN, use_lock=True)
    finally:
        # Ensure the executor is shut down properly
        executor.shutdown(wait=True)
        sys.stdout.write('\n')

    del executor

    # If there are more than 5 sources
    if output.shape[1] > config['NUM_STD_MIN']:
        x = np.nanmax(output[:, :, 2], axis=0)
        idx = np.where(x >= np.nanmean(x))[0]
        if len(idx) <= config['NUM_STD_MIN']:
            output = output[:, :int(config['NUM_STD_MIN'] + 1), :]
        else:
            output = output[:, idx, :]

    max_snr_idx = np.nanargmax(np.nanmean(output[:, :, 2], axis=1))
    max_snr_aprad = float(rapers[max_snr_idx])
    optimum_aprad = max_snr_aprad * 1.25
    qlf_aprad = True
    if optimum_aprad > 3.:
        log.warning('Optimum radius seems high [%.1f x FWHM] '
                    '- setting to %.1f x FWHM' % (optimum_aprad, aper_rad))
        optimum_aprad = aper_rad
        qlf_aprad = False

    if not silent:
        log.info('    ==> best-fit aperture radius: %3.1f (FWHM)' % max_snr_aprad)
        log.info('    ==> optimum aperture radius (r x 1.25): %3.1f (FWHM)' % optimum_aprad)

    del image, std_cat
    gc.collect()
    # quit()
    return output, rapers, max_snr_aprad, optimum_aprad, qlf_aprad


def get_snr(flux_star: np.array, flux_bkg: np.array,
            t_exp: float, area: float,
            gain: float = 1., rdnoise: float = 0., dc: float = 0.):
    """

    Parameters
    ----------
    flux_star
    flux_bkg
    t_exp
    area
    gain
    rdnoise
    dc

    Returns
    -------

    """
    counts_source = flux_star * t_exp
    sky_shot = flux_bkg * t_exp * area

    read_noise = ((rdnoise ** 2) + (gain / 2.) ** 2) * area
    dark_noise = dc * t_exp * area

    snr = counts_source / np.sqrt(counts_source + sky_shot + read_noise + dark_noise)

    del flux_star, flux_bkg
    gc.collect()

    SNR_cleaned = [i if i > 0 else 0 for i in snr]

    SNR_err = np.array([2.5 * np.log10(1. + (1. / snr)) if snr > 0 else np.nan for snr in SNR_cleaned])

    return SNR_cleaned, SNR_err


def get_aper_photometry(image: np.ndarray,
                        src_pos: np.ndarray,
                        mask: np.array = None,
                        aper_mode: str = 'circ',
                        r_aper: float = 1.7,
                        r_in: float = 1.9,
                        r_out: float = 2.2,
                        width: float = 500.,
                        height: float = 3.4,
                        w_in: float = 520., w_out: float = 530.,
                        h_in: float = 4., h_out: float = 6.,
                        theta: float = 0.,
                        gain: float = 1.):
    """Measure aperture photometry"""

    if aper_mode == 'circ':
        sum_method = 'exact'
        annulus_aperture = CircularAnnulus(src_pos, r_in=r_in, r_out=r_out)
        aperture = CircularAperture(src_pos, r=r_aper)
    else:
        sum_method = 'center'
        annulus_aperture = RectangularAnnulus(src_pos, w_in=w_in, w_out=w_out,
                                              h_in=h_in, h_out=h_out, theta=theta)
        aperture = RectangularAperture(src_pos,
                                       w=width,
                                       h=height,
                                       theta=theta)

    # Get statistics with mask in mind
    bkg_stats = ApertureStats(image, annulus_aperture,
                              mask=mask,
                              # sigma_clip=sigclip,
                              sum_method=sum_method)

    # Background median and standard deviation of annulus
    bkg_median = bkg_stats.mean
    bkg_std = bkg_stats.std
    # print(bkg_std)
    # print(bkg_median, bkg_std)

    if np.any(image) < 0:
        img_error = None
    else:
        from photutils.datasets import make_noise_image
        image_masked = image * ~mask if mask is not None else image
        img_mean = np.nanmean(image_masked)
        if img_mean < 0:
            img_error = None
        else:

            img_error = make_noise_image(image.shape,
                                     distribution='poisson',
                                     mean=img_mean)

    # error = None
    # aper_stats = ApertureStats(image, aperture, mask=mask,
    #                            sigma_clip=None, sum_method='exact')
    # if aper_mode == 'rect':
    #
    #     fig = plt.figure(figsize=(10, 6))
    #
    #     gs = gridspec.GridSpec(1, 1)
    #     ax = fig.add_subplot(gs[0, 0])
    #     ax.imshow(image*~mask, origin='lower', interpolation='nearest')
    #     aperture.plot(axes=ax, **{'color': 'white', 'lw': 1.25, 'alpha': 0.75})
    #     annulus_aperture.plot(axes=ax, **{'color': 'red', 'lw': 1.25, 'alpha': 0.75})

    # Make sure to use the same area over which to perform the photometry
    aperture_area = aperture.area_overlap(data=image, mask=mask, method=sum_method)
    annulus_area = annulus_aperture.area_overlap(data=image, mask=mask, method=sum_method)
    # print(aperture_area, annulus_area)

    # Run the aperture photometry
    phot_table = aperture_photometry(image, aperture,
                                     mask=mask,
                                     error=img_error,
                                     method=sum_method)
    del mask, aperture, annulus_aperture

    # Rename pix position to make it consistent
    names = ('xcenter', 'ycenter')
    new_names = ('xcentroid', 'ycentroid')
    phot_table.rename_columns(names, new_names)

    # Convert to pandas dataframe
    phot_table = phot_table.to_pandas()

    # Get total background
    total_bkg = bkg_median * aperture_area

    # Subtract background
    phot_bkgsub = phot_table['aperture_sum'] - total_bkg
    # print(phot_bkgsub, total_bkg)

    phot_table['aperture_sum_bkgsub'] = phot_bkgsub
    phot_table.loc[phot_table['aperture_sum_bkgsub'] <= 0] = 0

    # Compute the flux error using DAOPHOT style
    flux_variance = phot_table['aperture_sum_bkgsub'].values
    if img_error is not None:
        flux_variance = phot_table['aperture_sum_err'].values ** 2

    bkg_var = (aperture_area * bkg_std ** 2.) * (1. + aperture_area / annulus_area)
    flux_error = (flux_variance / gain + bkg_var) ** 0.5
    phot_table['aperture_sum_bkgsub_err'] = flux_error

    del image, src_pos, bkg_stats, flux_variance
    gc.collect()

    return (phot_table['aperture_sum_bkgsub'].values,
            flux_error, phot_table, bkg_median, bkg_std, aperture_area)


def get_std_photometry(image, std_cat, src_pos, fwhm, config):
    """Perform aperture photometry on standard stars"""

    res_cat = std_cat.copy()
    aper_dict = dict(aper=config['APER_RAD_OPT'] * fwhm,
                     inf_aper=config['INF_APER_RAD_OPT'] * fwhm)
    # print(aper_dict)
    # Mask unwanted pixel
    mask = (image < 0)
    img_mask = config['img_mask']
    if img_mask is not None:
        mask |= img_mask

    for key, val in aper_dict.items():
        phot_res = get_aper_photometry(image, src_pos,
                                       mask=img_mask,
                                       r_aper=val,
                                       r_in=val + config['RSKYIN_OPT'] * fwhm,
                                       r_out=val + config['RSKYOUT_OPT'] * fwhm
                                       )

        aper_flux_counts, aper_flux_count_err, _, aper_bkg_counts, _, _ = phot_res
        res_cat[f'flux_counts_{key}'] = aper_flux_counts
        res_cat[f'flux_counts_err_{key}'] = aper_flux_count_err
        del phot_res

    del src_pos, image
    gc.collect()

    return res_cat


def get_sat_photometry(image, img_mask, src_pos, fwhm, width, theta, config):
    """Perform aperture photometry on satellite"""

    res_cat = {}
    aper_dict = dict(aper=config['APER_RAD_RECT'],
                     inf_aper=config['INF_APER_RAD_RECT'])

    # Mask unwanted pixel
    mask = (image < 0)
    if img_mask is not None:
        mask |= img_mask

    w_in = width + config['RSKYIN_OPT_RECT'] * fwhm
    w_out = width + config['RSKYOUT_OPT_RECT'] * fwhm
    h_in = config['RSKYIN_OPT_RECT']
    h_out = config['RSKYOUT_OPT_RECT']

    for key, val in aper_dict.items():
        phot_res = get_aper_photometry(image, src_pos,
                                       mask=img_mask,
                                       height=val * 2. * fwhm,
                                       width=width,
                                       h_in=(val + h_in) * 2. * fwhm,
                                       h_out=(val + h_out) * 2. * fwhm,
                                       w_in=val + w_in,
                                       w_out=val + w_out,
                                       theta=theta,
                                       aper_mode='rect')
        # print(phot_res)
        aper_flux_counts, aper_flux_count_err, _, aper_bkg_counts, _, _ = phot_res
        res_cat[f'flux_counts_{key}'] = aper_flux_counts
        res_cat[f'flux_counts_err_{key}'] = aper_flux_count_err
        del phot_res

    del src_pos, image
    gc.collect()

    return res_cat


def get_glint_photometry(image, valid_mask, src_pos, width, height, theta, config,
                         sum_method='exact'):
    """

    Parameters
    ----------
    image
    valid_mask
    src_pos
    width
    height
    theta
    config
    sum_method

    Returns
    -------

    """
    gain = 1.
    res_cat = {}
    negative_mask = (image < 0)
    sat_lim_mask = (image > config['sat_lim'])

    # Mask only invalid and negative values
    negative_mask |= valid_mask

    # Mask negative and saturated pixel
    sat_lim_mask |= negative_mask
    sat_lim_mask |= valid_mask

    mask_dict = dict(with_sat_src=negative_mask, no_sat_src=sat_lim_mask)

    for key, mask in mask_dict.items():
        res_cat[key] = {}
        aperture = RectangularAperture(src_pos,
                                       w=width,
                                       h=height,
                                       theta=theta)

        rect_mask = aperture.to_mask(method='center')
        mask_img = rect_mask.to_image(image.shape)
        mask_img = np.where(mask_img == 1., True, False)

        bkg_mean, bkg_median, bkg_std = sigma_clipped_stats(image, mask_img)

        # Make sure to use the same area over which to perform the photometry
        aperture_area = aperture.area_overlap(data=image, mask=mask, method=sum_method)
        annulus_area = aperture.area_overlap(data=image, mask=mask_img, method=sum_method)

        # Run the aperture photometry
        phot_table = aperture_photometry(image, aperture,
                                         mask=mask,
                                         method=sum_method)
        del aperture

        # Rename pix position to make it consistent
        names = ('xcenter', 'ycenter')
        new_names = ('xcentroid', 'ycentroid')
        phot_table.rename_columns(names, new_names)

        # Convert to pandas dataframe
        phot_table = phot_table.to_pandas()

        # Get total background
        total_bkg = bkg_median * aperture_area

        # Subtract background
        phot_bkgsub = phot_table['aperture_sum'] - total_bkg
        # print(phot_bkgsub, total_bkg)

        phot_table['aperture_sum_bkgsub'] = phot_bkgsub
        phot_table.loc[phot_table['aperture_sum_bkgsub'] <= 0] = 0

        # Compute the flux error using DAOPHOT style
        flux_variance = phot_table['aperture_sum_bkgsub'].values
        bkg_var = (aperture_area * bkg_std ** 2.) * (1. + aperture_area / annulus_area)
        flux_error = (flux_variance / gain + bkg_var) ** 0.5

        res_cat[key]['flux_counts'] = phot_table['aperture_sum'].values[0]
        res_cat[key]['flux_counts_err'] = flux_error[0]
        res_cat[key]['area_pix'] = aperture_area

        del phot_table, mask_img
    del image, valid_mask
    return res_cat


def order_shift(x):
    """
    Get the order of magnitude of an array.
    Use to shift the array to between 0-10:
    This function uses the following equation to find the order of magnitude of an array.

    ..math::

       order_of_mag (x) = 10^{MAX(FLOOR(Log_{10}(x)))}

    :param x: Array of values
    :type x: float
    :return: Order of magnitude
    :rtype: float

    """
    idx = (np.isnan(x)) | (np.isinf(x)) | (x <= 0)

    order_of_mag = 10 ** np.nanmax(np.floor(np.log10(x[~idx])))

    return order_of_mag



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         base_config.py
# Purpose:      Base configuration file for reduction, astrometric calibration
#               and satellite trail analysis
#
#
# Author:       p4adch (cadam)
#
# Created:      04/15/2022
# Copyright:    (c) p4adch 2010-
#
# -----------------------------------------------------------------------------

""" Modules """
import gc
import os
import sys
import logging
import warnings
from dateutil.parser import parse
from packaging import version
from urllib import request
import numpy as np
from colorlog import ColoredFormatter
from astropy import wcs
from astropy.io import fits
from astropy.utils.exceptions import (AstropyUserWarning, AstropyWarning)

from threading import Lock

# from .telescope_conf import *
from .version import __version__

""" Meta-info """
__author__ = "Christian Adam"
__copyright__ = 'Copyright 2021, UA, LEOSat observations'
__credits__ = ["Christian Adam, Eduardo Unda-Sanzana, Jeremy Tregloan-Reed"]
__license__ = "Free"
__maintainer__ = "Christian Adam"
__email__ = "christian.adam84@gmail.com"
__status__ = "Production"

# Logging and console output
LOG_LEVEL = logging.INFO
LOGFORMAT = '[%(log_color)s%(levelname)8s%(reset)s] %(log_color)s%(message)s%(reset)s'
FORMATTER = ColoredFormatter(LOGFORMAT)


def load_warnings():
    warnings.simplefilter(action="ignore", category=RuntimeWarning)
    warnings.simplefilter(action='ignore', category=UserWarning)
    warnings.simplefilter(action='ignore', category=AstropyWarning)
    warnings.simplefilter(action='ignore', category=AstropyUserWarning)
    warnings.filterwarnings(action='ignore', category=wcs.FITSFixedWarning)
    warnings.filterwarnings(action='ignore', category=fits.column.VerifyWarning)
    warnings.filterwarnings(action='ignore', category=fits.card.VerifyWarning)

    # the following warning gets cast by Gaia query: XXX.convert_unit_to(u.deg)
    warnings.filterwarnings(action='ignore', category=np.ma.core.MaskedArrayFutureWarning)
    warnings.filterwarnings(action='ignore', category=UserWarning)
    warnings.filterwarnings(action='ignore', category=FutureWarning)

    # numpy warnings
    np.seterr(divide='ignore', invalid='ignore')
    np.errstate(invalid='ignore')


# Lock for thread-safe printing
print_lock = Lock()

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), '../..'))


# Color definitions for console output
class BCOLORS:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[32m'
    WARNING = '\033[93m'
    PASS = '\033[92m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


# Color strings for console output
pass_str = BCOLORS.PASS + "SUCCESSFUL" + BCOLORS.ENDC
fail_str = BCOLORS.FAIL + "FAILED" + BCOLORS.ENDC

# List of potential FITS header keywords that can be used to identify the instrument
# Any unique header keyword serves as an identifier
INSTRUMENT_KEYS = ['PPINSTRU', 'LCAMMOD', 'INSTRUME', 'INSTRUM',
                   'TELESCOP', 'FLI', 'FPA', 'CAM_NAME']


FRMT_FS = "%Y-%m-%dT%H:%M:%S.%f"
FRMT = "%Y-%m-%dT%H:%M:%S"


def has_fractional_seconds(time_string):
    # Parse the time string into a datetime object
    dt = parse(time_string)

    # Check if the datetime object has fractional seconds
    if dt.microsecond > 0:
        return FRMT_FS
    else:
        return FRMT


# Clean up function to delete variables and free memory
def clean_up(*args):
    """Delete variables and free memory."""

    for _ in args:
        del _

    gc.collect()


# Default number of decimal places used in rounding
ROUND_DECIMAL = 5

IMAGETYP_LIGHT = 'science'
IMAGETYP_BIAS = 'bias'
IMAGETYP_DARK = 'dark'
IMAGETYP_FLAT = 'flat'

IMAGETYP_REDUCE_ORDER = np.array(['bias', 'darks', 'flats'])
IMAGETYP_COMBOS = {'bias': "zero|bias|bias frame",
                   'darks': "dark|dark frame",
                   'flats': "flat|flat frame|dflat|sflat",
                   'light': "science|light|object|sky|light frame"}

BINNING_DKEYS = ('BINX', 'BINY')

# Default result table column names.
DEF_RES_TBL_COL_NAMES = ['File', 'Object', 'Sat-Name', 'AltID', 'UniqueID',
                         'Instrument', 'HDU_idx', 'DetPosID',
                         'Telescope', 'RA', 'DEC', 'Date-Obs', 'Filter', 'ExpTime',
                         'Airmass', 'Binning',
                         'Obs-Start', 'Obs-Mid', 'Obs-Stop',
                         'HasTrail', 'NTrail', 'HasRef', 'NRef',
                         'UT Date', 'UT time',
                         'SatLon', 'SatLat', 'SatAlt',
                         'SatAz', 'SatElev', 'SatRA', 'SatDEC',
                         'SunRA', 'SunDEC', 'SunZenithAngle', 'SatAngularSpeed', 'e_SatAngularSpeed',
                         'ObsSatRange',
                         'ToD', 'e_ToD',
                         'ObsTrailLength', 'e_ObsTrailLength', 'EstTrailLength', 'e_EstTrailLength',
                         'SunSatAng', 'SunPhaseAng', 'SunIncAng', 'ObsAng',
                         'ObsMag', 'e_ObsMag', 'EstMag', 'e_EstMag',
                         'EstScaleMag', 'e_EstScaleMag',
                         'HasZp', 'MagZp',
                         'ObsMag_zp', 'e_ObsMag_zp',
                         'EstMag_zp', 'e_EstMag_zp',
                         'EstScaleMag_zp', 'e_EstScaleMag_zp',
                         'SunAzAng', 'SunElevAng',
                         'FluxScale', 'MagScale', 'MagCorrect', 'e_MagCorrect', 'MagCorrectSat',
                         'dt_tle-obs',
                         'TrailCX', 'e_TrailCX', 'TrailCY', 'e_TrailCY',
                         'TrailCRA', 'e_TrailCRA', 'TrailCDEC', 'e_TrailCDEC',
                         'TrailANG', 'e_TrailANG', 'OptAperHeight',
                         'CalRA', 'CalDEC', 'CentX', 'e_CentX', 'CentY', 'e_CentY', 'PixScale', 'DetRotAng', 'FWHM', 'e_FWHM',
                         'bias_cor', 'dark_cor', 'flat_cor', 'WCS_cal', 'mag_conv', 'QlfAperRad', 'TrailDetMethod']

# To be implemented. Default result table column units.
DEF_RES_TBL_COL_UNITS = ['', '', '', '', '',
                         '', '', '',
                         '', 'hh:mm:ss.sss', 'deg:mm:ss.sss', '', '', 's',
                         '', '',
                         'hh:mm:ss.sss', 'hh:mm:ss.sss', 'hh:mm:ss.sss',
                         '', '', '', '',
                         'YYYY-MM-DD', 'hh:mm:ss.sss',
                         'deg', 'deg', 'km',
                         'deg', 'deg', 'hh:mm:ss.sss', 'deg',
                         'hr', 'deg', 'deg', 'arcsec/s', 'arcsec/s',
                         'km',
                         's', 's',
                         'arcsec', 'arcsec', 'arcsec', 'arcsec',
                         'deg', 'deg', 'deg', 'deg',
                         'mag', 'mag', 'mag', 'mag',
                         'mag', 'mag',
                         '', '',
                         'mag', 'mag',
                         'mag', 'mag',
                         'mag', 'mag',
                         'deg', 'deg',
                         '', '', 'mag', 'mag', 'mag', 's',
                         'px', 'px', 'px', 'px',
                         'deg', 'arcsec', 'deg', 'arcsec',
                         'deg', 'arcsec', 'px',
                         'deg', 'deg', 'px', 'px', 'px', 'px', 'arcsec/px', 'deg', 'px', 'px',
                         '', '', '', '', '', '', '']

# Default visibility file table column names.
DEF_VIS_TBL_COL_NAMES = ['ID', 'AltID', 'UniqueID', 'UT Date', 'UT time',
                         'SatLon', 'SatLat', 'SatAlt',
                         'SatAz', 'SatElev', 'SatRA', 'SatDEC',
                         'SunRA', 'SunDEC', 'SunZenithAngle',
                         'SatAngularSpeed']

# Translations between table and fits header keywords
DEF_KEY_TRANSLATIONS = {
    'Object': ['OBJECT', 'BLKNM'],
    'Instrument': ['INSTRUME', 'INSTRUM'],
    'Telescope': ['TELESCOP', 'OBSERVAT'],
    'HDU_idx': ['HDU_IDX'],
    'DetPosID': ['DETPOS'],
    'Filter': ['FILTER', 'BAND'],
    'ExpTime': ['EXPTIME'],
    'RA': ['RA', 'OBSRA', 'OBJRA', 'OBJCTRA', 'STRRQRA', 'OBJ-RA'],
    'DEC': ['DEC', 'OBSDEC', 'OBJDEC', 'OBJCTDEC', 'STRRQDE', 'OBJ-DEC'],
    'CalRA': ['CRVAL1'],
    'CalDEC': ['CRVAL2'],
    'CentX': ['CRPIX1'],
    'e_CentX': ['POSERRX'],
    'CentY': ['CRPIX2'],
    'e_CentY': ['POSERRY'],
    'PixScale': ['PIXSCALE'],
    'DetRotAng': ['DETROTANG'],
    'FWHM': ['FWHM'],
    'e_FWHM': ['FWHMERR'],
    'Airmass': ['AIRMASS', 'STROBAM'],
    'Date-Obs': ['DATE-OBS', 'DATE-BEG'],
    'Binning': ['BINNING'],
    'bias_cor': ['BIAS_COR'],
    'dark_cor': ['DARK_COR'],
    'flat_cor': ['FLAT_COR'],
    'WCS_cal': ['AST_CAL', 'WCSCAL', 'WCS-STAT'],
    'HasTrail': ['HASTRAIL'],
    'NTrail': ['NTRAIL']
}

# List of available catalogs for photometry
ALLCATALOGS = ['GAIADR1', 'GAIADR2', 'GAIADR3', 'GAIAEDR3',
               '2MASS', 'PS1DR1', 'PS1DR2',
               'GSC241', 'GSC242', 'GSC243']

# Astroquery catalogs
DEF_ASTROQUERY_CATALOGS = {
    # 'GAIADR3': 'gaiadr3.gaia_source',
    'GAIADR3': 'gaiadr3.gaia_source_lite',
    'GSC242': 'I/353/gsc242'
}

# Dictionary of supported astrometric catalogs with column name translations.
# These are the minimum columns necessary for alignment to work properly while
# taking into account all available information from the catalogs.
# Blank column names indicate the catalog does not have that column,
# and will set to None in the output table.
SUPPORTED_CATALOGS = {
    'GAIADR3': {'RA': 'ra',
                # 'RA_error': 'ra_error',
                'DEC': 'dec',
                # 'DEC_error': 'dec_error',
                'pmra': 'pmra', 'pmra_error': 'pmra_error',
                'pmdec': 'pmdec', 'pmdec_error': 'pmdec_error',
                'mag': 'phot_g_mean_mag', 'objID': 'SOURCE_ID',
                # 'epoch': 'ref_epoch'
                },
    'GAIAEDR3': {'RA': 'ra', 'RA_error': 'ra_error',
                 'DEC': 'dec', 'DEC_error': 'dec_error',
                 'pmra': 'pmra', 'pmra_error': 'pmra_error',
                 'pmdec': 'pmdec', 'pmdec_error': 'pmdec_error',
                 'mag': 'mag', 'objID': 'source_id', 'epoch': 'epoch'},
    'GAIADR2': {'RA': 'ra', 'RA_error': 'ra_error',
                'DEC': 'dec', 'DEC_error': 'dec_error',
                'pmra': 'pmra', 'pmra_error': 'pmra_error',
                'pmdec': 'pmdec', 'pmdec_error': 'pmdec_error',
                'mag': 'mag', 'objID': 'objID', 'epoch': 'epoch'},
    'GAIADR1': {'RA': 'ra', 'RA_error': 'ra_error',
                'DEC': 'dec', 'DEC_error': 'dec_error',
                'pmra': 'pmra', 'pmra_error': 'pmra_error',
                'pmdec': 'pmdec', 'pmdec_error': 'pmdec_error',
                'mag': 'mag', 'objID': 'objID', 'epoch': 'epoch'},
    '2MASS': {'RA': 'ra', 'RA_error': '',
              'DEC': 'dec', 'DEC_error': '',
              'pmra': '', 'pmra_error': '',
              'pmdec': '', 'pmdec_error': '',
              'mag': 'mag', 'objID': 'objid', 'epoch': 'jdate'},
    'PS1DR2': {'RA': 'ra', 'RA_error': 'raMeanErr',
               'DEC': 'dec', 'DEC_error': 'decMeanErr',
               'pmra': '', 'pmra_error': '',
               'pmdec': '', 'pmdec_error': '',
               'mag': 'mag', 'objID': 'objid', 'epoch': 'epochMean'},
    'PS1DR1': {'RA': 'ra', 'RA_error': 'RaMeanErr',
               'DEC': 'dec', 'DEC_error': 'DecMeanErr',
               'pmra': '', 'pmra_error': '',
               'pmdec': '', 'pmdec_error': '',
               'mag': 'mag', 'objID': 'objID', 'epoch': ''},
    'GSC243': {'RA': 'ra', 'RA_error': 'raErr',
               'DEC': 'dec', 'DEC_error': 'decErr',
               'pmra': 'rapm', 'pmra_error': 'rapmErr',
               'pmdec': 'decpm', 'pmdec_error': 'decpmErr',
               'mag': 'mag', 'objID': 'objID', 'epoch': 'epoch'},
    'GSC242': {'RA': 'ra', 'RA_error': 'raErr',
               'DEC': 'dec', 'DEC_error': 'decErr',
               'pmra': 'rapm', 'pmra_error': 'rapmErr',
               'pmdec': 'decpm', 'pmdec_error': 'decpmErr',
               'mag': 'mag', 'objID': 'objID', 'epoch': 'epoch'},
    'GSC241': {'RA': 'ra', 'RA_error': 'raErr',
               'DEC': 'dec', 'DEC_error': 'decErr',
               'pmra': 'rapm', 'pmra_error': 'rapmErr',
               'pmdec': 'decpm', 'pmdec_error': 'decpmErr',
               'mag': 'mag', 'objID': 'objID', 'epoch': 'epoch'}
}

# Dictionary of supported filter bands available in the supported catalogs
SUPPORTED_BANDS = {
    "U": ["GSC243", "GSC242"], "B": ["GSC243", "GSC242"],
    "V": ["GSC243", "GSC242"], "R": ["GSC243", "GSC242"], "I": ["GSC243", "GSC242"],
    "J": ["2MASS"], "H": ["2MASS"], "K": ["2MASS"], "Ks": ["2MASS"],
    "g": ["GSC243", "GSC242"], "r": ["GSC243", "GSC242"], "i": ["GSC243", "GSC242"], "u": ["GSC243", "GSC242"],
    "w": ["GSC243", "GSC242"], "y": ["GSC243", "GSC242"], "z": ["GSC243", "GSC242"]
}

# Dictionary of translations for the filter in the supported catalogs
CATALOG_FILTER_EXT = {
    'GAIAEDR3': None,
    'GAIADR3': None,
    'GAIADR2': None,
    'GAIADR1': None,
    '2MASS': {"J": {'Prim': [['j_m', 'j_cmsig']], 'Alt': None},
              "H": {'Prim': [['h_m', 'h_cmsig']], 'Alt': None},
              "K": {'Prim': [['k_m', 'k_cmsig']], 'Alt': None},
              "Ks": {'Prim': [['k_m', 'k_cmsig']], 'Alt': None}},
    'PS1DR2': {"u": {'Prim': [['SDSSuMag', 'SDSSuMagErr']], 'Alt': None},
               "g": {'Prim': [['SDSSgMag', 'SDSSgMagErr']], 'Alt': None},
               "r": {'Prim': [['SDSSrMag', 'SDSSrMagErr']], 'Alt': None},
               "i": {'Prim': [['SDSSiMag', 'SDSSiMagErr']], 'Alt': None},
               "z": {'Prim': [['SDSSzMag', 'SDSSzMagErr']], 'Alt': None},
               "y": {'Prim': [['PS1yMag', 'PS1ymagErr']], 'Alt': None}},
    'PS1DR1': None,
    'GSC243': {"U": {'Prim': [['Umag', 'UmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "B": {'Prim': [['Bmag', 'BmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "V": {'Prim': [['Vmag', 'VmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "R": {'Prim': [['Rmag', 'RmagErr']],
                     'Alt': [['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr'],
                             ['SDSSiMag', 'SDSSiMagErr']]},
               "I": {'Prim': [['Imag', 'ImagErr']],
                     'Alt': [['SDSSrMag', 'SDSSrMagErr'],
                             ['SDSSiMag', 'SDSSiMagErr'],
                             ['SDSSzMag', 'SDSSzMagErr']]},
               "g": {'Prim': [['SDSSgMag', 'SDSSgMagErr']], 'Alt': None},
               "r": {'Prim': [['SDSSrMag', 'SDSSrMagErr']], 'Alt': None},
               "i": {'Prim': [['SDSSiMag', 'SDSSiMagErr']], 'Alt': None},
               "u": {'Prim': [['SDSSuMag', 'SDSSuMagErr']], 'Alt': None},
               "w": {'Prim': [['SDSSgMag', 'SDSSgMagErr'],
                              ['SDSSrMag', 'SDSSrMagErr']], 'Alt': None},
               "z": {'Prim': [['SDSSzMag', 'SDSSzMagErr']], 'Alt': None}},
    'GSC242': {"U": {'Prim': [['Umag', 'UmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "B": {'Prim': [['Bmag', 'BmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "V": {'Prim': [['Vmag', 'VmagErr']],
                     'Alt': [['SDSSuMag', 'SDSSuMagErr'],
                             ['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr']]},
               "R": {'Prim': [['Rmag', 'RmagErr']],
                     'Alt': [['SDSSgMag', 'SDSSgMagErr'],
                             ['SDSSrMag', 'SDSSrMagErr'],
                             ['SDSSiMag', 'SDSSiMagErr']]},
               "I": {'Prim': [['Imag', 'ImagErr']],
                     'Alt': [['SDSSrMag', 'SDSSrMagErr'],
                             ['SDSSiMag', 'SDSSiMagErr'],
                             ['SDSSzMag', 'SDSSzMagErr']]},
               "g": {'Prim': [['SDSSgMag', 'SDSSgMagErr']], 'Alt': None},
               "r": {'Prim': [['SDSSrMag', 'SDSSrMagErr']], 'Alt': None},
               "i": {'Prim': [['SDSSiMag', 'SDSSiMagErr']], 'Alt': None},
               "w": {'Prim': [['SDSSgMag', 'SDSSgMagErr'],
                              ['SDSSrMag', 'SDSSrMagErr']], 'Alt': None},
               "z": {'Prim': [['SDSSzMag', 'SDSSzMagErr']], 'Alt': None}},
    'GSC241': None
}

# Color conversion factors for transformation from SSDS mag to BVRI following Lupton (2005)
CONV_SSDS_BVRI = {'Bmag': [[-0.8116, 0.1313, 0.0095],
                           [0.3130, 0.2271, 0.0107]],
                  'Vmag': [[-0.2906, 0.0885, 0.0129],
                           [-0.5784, -0.0038, 0.0054]],
                  'Rmag': [[-0.1837, -0.0971, 0.0106],
                           [-0.2936, -0.1439, 0.0072]],
                  'Imag': [[-1.2444, -0.3820, 0.0078],
                           [-0.3780, -0.3974, 0.0063]]}

# Dictionary of satellite properties
SATELLITE_PROPERTIES = {
    'ONEWEB': {
        'identifiers': ['ONEWEB', 'OW'],
        'format_number': lambda x: f"ONEWEB-{int(x):04d}",
        'orbit_height': 1200.0
    },
    'STARLINK': {
        'identifiers': ['STARLINK'],
        'format_number': lambda x: f"STARLINK-{int(x):04d}" if len(x) == 4 else f"STARLINK-{int(x):05d}",
        'orbit_height': 550.0
    },
    'BLUEWALKER': {
        'identifiers': ['BLUEWALKER', 'BW'],
        'format_number': lambda x: f"BLUEWALKER-{int(x)}",
        'orbit_height': 500.0
    },
    'KUIPER': {
        'identifiers': ['KUIPER'],
        'format_number': lambda x: f"KUIPER-P{int(x)}",
        'orbit_height': 630.0
    },
    'SPACEMOBILE': {
        'identifiers': ['SPACEMOBILE'],
        'format_number': lambda x: f"SPACEMOBILE-{int(x):03d}",
        'orbit_height': 550.0
    }
}


def get_nominal_orbit_altitude(sat_id: str) -> float:
    """
    Get the nominal orbit altitude for a given satellite ID.

    Args:
        sat_id (str): The satellite ID.

    Returns:
        float: The nominal orbit altitude in kilometers.
    """
    for key, props in SATELLITE_PROPERTIES.items():
        if key in sat_id:
            return props['orbit_height']
    # Default to ONEWEB orbit height if no match is found
    return SATELLITE_PROPERTIES['ONEWEB']['orbit_height']


REARTH_EQU = 6378.137  # Radius at sea level at the equator in km
REARTH_POL = 6356.752  # Radius at poles in km

AU_TO_KM = 149597892.  # km/au

# URL configuration for the photometry pipeline
ASTROMETRIC_CAT_ENVVAR = "ASTROMETRIC_CATALOG_URL"
DEF_CAT_URL = 'https://gsss.stsci.edu/webservices'

if ASTROMETRIC_CAT_ENVVAR in os.environ:
    SERVICELOCATION = os.environ[ASTROMETRIC_CAT_ENVVAR]
else:
    SERVICELOCATION = DEF_CAT_URL


def check_version(log):
    """Crosscheck the current version with the GitHub version"""
    url_github_version = 'https://raw.githubusercontent.com/CLEOsat-group/leosatpy/master/leosatpy/utils/version.py'

    log.info("> Crosscheck local version with GitHub")

    try:
        version_string = str(request.urlopen(url_github_version).readlines()[0], 'utf-8').strip()
        version_github = version_string[version_string.rfind('=') + 1:].replace(' ', '').replace('\'', '')
    except (Exception, ValueError):
        version_github = ''

    if version_github and version.parse(__version__) < version.parse(version_github):
        s = ' ' * 15
        log.warning(f'  ==> A newer version of LEOSatpy is available (v{__version__} --> v{version_github}).\n '
                    f'{s} Please consider updating LEOSatpy by typing "pip install leosatpy --upgrade" in the terminal.')
        ask_str = BCOLORS.OKGREEN + "[    INFO]   >> Press Enter to continue <<" + BCOLORS.ENDC
        input(ask_str)
    elif version_github and version.parse(__version__) > version.parse(version_github):
        log.info("  ==> You are using a newer version of LEOSatpy than the official release.")
    else:
        all_good = BCOLORS.PASS + "ALL GOOD" + BCOLORS.OKGREEN
        log.info(f"  ==> {all_good}: You are using the latest official release of LEOSatpy." + BCOLORS.ENDC)


def print_progress_bar(iteration, total, prefix='', length=50, fill='\u2588', color=None, use_lock=False):
    """
    Print a progress bar in the terminal with ANSI color formatting.

    Parameters
    ----------
    iteration : int
        Current iteration step.
    total : int
        Total number of iterations.
    prefix : str, optional
        Prefix to display before the progress bar.
    length : int, optional
        Length of the progress bar in characters.
    fill : str, optional
        Character used to fill the progress bar.
    color : str, optional
        ANSI color code to colorize the progress bar.
    use_lock : bool, optional
        Whether to use a lock for printing (useful in multithreaded contexts).
    """

    # Ensure the percentage reaches 100% at the final iteration
    if iteration >= total:
        percent = "100.0"
        filled_length = length
    else:
        percent = "{0:.1f}".format(100 * (iteration / float(total)))
        filled_length = int(length * iteration // total)

    bar = fill * filled_length + '-' * (length - filled_length)

    color_prefix = color if color else ''
    reset_color = BCOLORS.ENDC if color else ''
    output = (f'\r[{color_prefix}PROGRESS{reset_color}]'
              f'{color_prefix}{prefix}{reset_color} |{bar}| '
              f'{color_prefix}{percent}%{reset_color}')

    if use_lock:
        with print_lock:
            sys.stdout.write(output)
            sys.stdout.flush()
    else:
        sys.stdout.write(output)
        sys.stdout.flush()
